import { Component, OnDestroy, OnInit } from "@angular/core";
import { FormBuilder, FormControl, FormGroup } from "@angular/forms";
import { Page } from "../../../../../../models/graphql/filter/page.model";
import { Sort, SortDirection } from "../../../../../../models/graphql/filter/sort.model";
import { Subject, throwError } from "rxjs";
import { Server } from "../../../../../../models/data/server.model";
import { UserDTO } from "../../../../../../models/auth/auth-user.dto";
import { Queries } from "../../../../../../models/graphql/queries.model";
import { AuthService } from "../../../../../../services/auth/auth.service";
import { SpinnerService } from "../../../../../../services/spinner/spinner.service";
import { ToastService } from "../../../../../../services/toast/toast.service";
import { ConfirmationDialogService } from "../../../../../../services/confirmation-dialog/confirmation-dialog.service";
import { TranslateService } from "@ngx-translate/core";
import { DataService } from "../../../../../../services/data/data.service";
import { SidebarService } from "../../../../../../services/sidebar/sidebar.service";
import { ActivatedRoute } from "@angular/router";
import { Query, UserListQuery, UserWarningQuery } from "../../../../../../models/graphql/query.model";
import { catchError, debounceTime, takeUntil } from "rxjs/operators";
import { LazyLoadEvent } from "primeng/api";
import { Table } from "primeng/table";
import { User } from "../../../../../../models/data/user.model";
import { UpdateUserMutationResult } from "../../../../../../models/graphql/result.model";
import { Mutations } from "../../../../../../models/graphql/mutations.model";
import { ComponentWithTable } from "../../../../../../base/component-with-table";
import { UserWarning, UserWarningFilter } from "../../../../../../models/data/user_warning.model";

@Component({
  selector: "app-user-warning",
  templateUrl: "./user-warning.component.html",
  styleUrls: ["./user-warning.component.scss"]
})
export class UserWarningComponent extends ComponentWithTable implements OnInit, OnDestroy {
  public userWarnings: UserWarning[] = [];
  public users: User[] = [];
  public loading = true;

  public filterForm!: FormGroup<{
    id: FormControl<number | null>,
    description: FormControl<string | null>,
    user: FormControl<string | null>,
    author: FormControl<string | null>,
  }>;

  public filter: UserWarningFilter = {};
  public page: Page = {
    pageSize: undefined,
    pageIndex: undefined
  };
  public sort: Sort = {
    sortColumn: undefined,
    sortDirection: undefined
  };

  public totalRecords: number = 0;

  public clonedUserWarning: { [s: string]: UserWarning; } = {};

  private unsubscriber = new Subject<void>();
  private server: Server = {};
  public user: UserDTO | null = null;

  query: string = Queries.userWarningsQuery;

  public constructor(
    private authService: AuthService,
    private spinner: SpinnerService,
    private toastService: ToastService,
    private confirmDialog: ConfirmationDialogService,
    private fb: FormBuilder,
    private translate: TranslateService,
    private data: DataService,
    private sidebar: SidebarService,
    private route: ActivatedRoute) {
    super("UserWarning", ["id", "description", "member", "author"],
      (oldElement: UserWarning, newElement: UserWarning) => {
        return oldElement.id === newElement.id;
      });
  }

  public ngOnInit(): void {
    this.loading = true;
    this.data.getServerFromRoute(this.route).then(async server => {
      this.server = server;
      let authUser = await this.authService.getLoggedInUser();
      this.user = authUser?.users?.find(u => u.server == this.server.id) ?? null;
      this.setFilterForm();
    });
  }

  public ngOnDestroy(): void {
    this.unsubscriber.next();
    this.unsubscriber.complete();
  }

  public loadNextPage(): void {
    this.data.query<UserWarningQuery>(Queries.userWarningsQuery, {
        serverId: this.server.id, filter: this.filter, page: this.page, sort: this.sort
      },
      (data: Query) => {
        return data.servers[0];
      }
    ).subscribe(data => {
      this.totalRecords = data.userWarningCount;
      this.userWarnings = data.userWarnings;

      this.data.query<UserListQuery>(Queries.liteUsersQuery, {
          serverId: this.server.id, filter: {
            leftServer: false
          }
        },
        (x: { servers: Server[] }) => {
          return x.servers[0];
        }
      ).subscribe(data => {
        this.users = data.users;

        this.spinner.hideSpinner();
        this.loading = false;
      });
    });
  }

  public setFilterForm(): void {
    this.filterForm = this.fb.group({
      id: new FormControl<number | null>(null),
      description: new FormControl<string | null>(null),
      user: new FormControl<string | null>(null),
      author: new FormControl<string | null>(null)
    });

    this.filterForm.valueChanges.pipe(
      takeUntil(this.unsubscriber),
      debounceTime(600)
    ).subscribe(changes => {
      if (changes.id) {
        this.filter.id = changes.id;
      } else {
        this.filter.id = undefined;
      }

      if (this.page.pageSize)
        this.page.pageSize = 10;

      if (this.page.pageIndex)
        this.page.pageIndex = 0;

      this.loadNextPage();
    });
  }

  public newUserWarningTemplate: UserWarning = {
    createdAt: "",
    modifiedAt: ""
  };

  public nextPage(event: LazyLoadEvent): void {
    this.page.pageSize = event.rows ?? 0;
    if (event.first != null && event.rows != null)
      this.page.pageIndex = event.first / event.rows;
    this.sort.sortColumn = event.sortField ?? undefined;
    this.sort.sortDirection = event.sortOrder === 1 ? SortDirection.ASC : event.sortOrder === -1 ? SortDirection.DESC : SortDirection.ASC;

    this.loadNextPage();
  }

  public resetFilters(): void {
    this.filterForm.reset();
  }

  public onRowEditInit(table: Table, user: User, index: number): void {
    this.clonedUserWarning[index] = { ...user };
  }

  public override onRowEditSave(newUserWarning: UserWarning, index: number): void {
    if (this.isEditingNew && JSON.stringify(newUserWarning) === JSON.stringify(this.newUserWarningTemplate)) {
      this.isEditingNew = false;
      this.userWarnings.splice(index, 1);
      return;
    }

    if (!newUserWarning.id && !this.isEditingNew) {
      return;
    }

    this.updateUserByWarning(newUserWarning.user?.id ?? 0, this.userWarnings.filter(uw => uw.user?.id == newUserWarning.user?.id));
    this.isEditingNew = false;
  }

  public onRowEditCancel(index: number): void {
    if (this.isEditingNew) {
      this.userWarnings.splice(index, 1);
      delete this.clonedUserWarning[index];
      this.isEditingNew = false;
      return;
    }

    this.userWarnings[index] = this.clonedUserWarning[index];
    delete this.clonedUserWarning[index];
  }

  public addUserWarning(table: Table): void {
    const newUserWarning = JSON.parse(JSON.stringify(this.newUserWarningTemplate));

    this.userWarnings = [newUserWarning, ...this.userWarnings];

    table.initRowEdit(newUserWarning);

    const index = this.userWarnings.findIndex(l => l.id == newUserWarning.id);
    this.onRowEditInit(table, newUserWarning, index);

    this.isEditingNew = true;
  }

  public deleteUserWarning(userWarning: UserWarning): void {
    this.userWarnings.splice(this.userWarnings.findIndex(uw => uw.id == userWarning.id), 1);
    this.updateUserByWarning(userWarning.user?.id ?? 0, this.userWarnings.filter(uw => uw.user?.id == userWarning.user?.id));
  }

  public updateUserByWarning(userId: number, userWarnings: UserWarning[]) {
    this.spinner.showSpinner();
    this.data.mutation<UpdateUserMutationResult>(Mutations.updateUser, {
        id: userId,
        userWarnings: userWarnings?.map(userWarning => {
          return {
            id: userWarning.id,
            user: userWarning.user?.id ?? undefined,
            description: userWarning.description,
            author: userWarning.author?.id ?? undefined
          };
        })
      }
    ).pipe(catchError(err => {
      this.spinner.hideSpinner();
      return throwError(err);
    })).subscribe(_ => {
      this.spinner.hideSpinner();
      this.toastService.success(this.translate.instant("view.server.user_warning.message.user_warning_updated"), this.translate.instant("view.server.user_warning.message.user_warning_updated"));
      this.loadNextPage();
    });
  }
}
